/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "Generator.hh"
#include "FrWriter.hh"
#include "SigFlag.hh"
#include "TrigPar.hh"
#include <string>
#include <map>

/**  \section dmtgen_intro Introduction
  *
  *  %DMTGen is a data simulation program written primarily for use in testing
  *  DMT code. A %DMTGen run produces a specified set of channels in data 
  *  frames. Each channels is generated from the combination of one or more
  *  signals as defined by the \c Source configuration comment (see
  *  \ref dmtgen_conf_srce) and produced by data source classes based on 
  *  the DataSource class. 
  *  The data sources can either read pre-recorded data from frames (or xml 
  *  in the future) or generate new data according to analytical or random 
  *  functions. Generated data can be either continuous or transient. 
  *  Transient data sources are generated at either a fixed time, at random 
  *  times or at evenly spaced times. Simulated data structures and detector 
  *  structures may also be included in the output frames.
  *
  *  \section dmtgen_run Running DMTGen
  *
  *  The %DMTGen command line syntax is as follows
  *
  *  \verbatim
     DMTGen -conf <c-file> [-realtime] [-debug <dbg-lvl>]
     \endverbatim
  *  Where \c \<c-file\> is a configuration file as described below in 
  *  \ref dmtgen_conf. If \c -realtime is specified, the the generator is 
  *  run in real-time mode, in which case the frames are written immediately 
  *  after the end time of the frame.
  *
  *  \section dmtgen_conf Configuring a DMTGen Generation Run
  *
  *  A %DMTGen generation run is specified by a configuration file containing:
  *   - Generation control parameters
  *   - Filter descriptions.
  *   - Data source descriptions
  *   - %Channel descriptions.
  *
  *  Configuration files consist of one or more configuration commands. 
  *  Each command is a series of white-space delimited words. The command
  *  ends with a pound sign ("#") or at the end of a line. If the command 
  *  can not be contained on a single line, it may be continued on the next
  *  line by ending the first through penultimate line(s) with a backslash
  *  ("\"). Comment lines begin with a pound sign ("#").
  *
  *  \subsection dmtgen_conf_incl Including Configuration Files
  *
  *  Configuration files may be included into other configurations with an 
  *  include statement. The include statement syntax is as following syntax:
  *  \verbatim
     Include  <file-name>
     \endverbatim
  *
  *  \subsection dmtgen_conf_par Generation Control Parameters
  *
  *  These parameters control various aspects of a generation run such as 
  *  length of the run, the output files, etc. The control parameters are 
  *  specified as follows:
  *  \verbatim 
     Parameter  <param-name>  <param-value>
     \endverbatim
  *  The currently defined parameters include:
  *
  *  <table>
  *  <tr><td>Parameter </td>
  *      <td> default </td>
  *      <td> Contents </td>
  *  </tr>
  *  <tr><td>\c Compression </td>
  *      <td> none </td>
  *      <td> Default compression scheme ("", "gzip", "zero-suppress-2", 
  *           "zero-suppress-4", "zero-suppress-or-gzip") [""]</td>
  *  </tr>
  *  <tr><td>\c EndGPS </td>
  *      <td> - </td>
  *       <td> End GPS time</td>
  *  </tr>
  *  <tr><td>\c FilePrefix </td>
  *      <td> %DMTGen </td>
  *      <td> Frame file names are \c \<prefix\>-\<start-gps\>-\<len\>.gwf</td>
  *  </tr>
  *  <tr><td>\c FrameLength </td>
  *      <td> 16 </td>
  *      <td> Length of the frame files (in seconds)</td>
  *  </tr>
  *  <tr><td>\c GenerStride </td>
  *      <td> 4.0 </td>
  *      <td> Number of seconds of data generated at once.</td>
  *  </tr>
  *  <tr><td>\c Identifier </td>
  *      <td> "DMT generation run" </td>
  *      <td> The identifier string written to the frame headers</td>
  *  </tr>
  *  <tr><td>\c OutputDirectory </td>
  *      <td> "" </td>
  *      <td> The directory to which output frames are written</td>
  *  </tr>
  *  <tr><td>\c RandomSeed </td>
  *      <td> none </td>
  *      <td> Set the random number generator seed. If a random state file
  *           is specfied in addition to the seed number, the state
  *           file will be used (see SaveRndmState parameter).</td>
  *  </tr>
  *  <tr><td>\c RealTime </td>
  *      <td> 0 </td>
  *      <td> Generate data in real-time. The generated data are labeled 
  *           with the current time and are not written until the end of
  *           frame time.
  *      </td>
  *  </tr>
  *  <tr><td>\c RunID </td>
  *      <td> -666 </td>
  *      <td> The generation run number (must be negative)</td>
  *  </tr>
  *  <tr><td>\c SaveRndmState </td>
  *      <td> none </td>
  *      <td> File to/from which the random number generator state is 
  *           saved/restored</td>
  *  </tr>
  *  <tr><td>\c StartGPS </td>
  *      <td> - </td>
  *      <td> Starting GPS time</td>
  *  </tr>
  *  </table>
  *
  *  \subsection dmtgen_conf_filt Filter Descriptions
  *
  *  Data from each source may be individually filtered. The data may also be 
  *  filtered before and after summation of the various sources that combine 
  *  to form a channel. The syntax for defining a filter is:
  *  \verbatim
     Filter  <filter-name>  <filter-type>  [ <p1> .... <pN> ]
     \endverbatim
  *  The following filter types have been defined
  *
  *  <table>
  *  <tr><td> Decimate </td>
  *      <td> Decimate data by the factor specified in \c P1. The optional
  *           second parameter gives the decimation method, either \e simple
  *           (default) which performs no anti-aliasing filtering, or 
  *           \e decimateby2 which uses the DMT DecimateBy2 class.</td>
  *  </tr>
  *  <tr><td>\c DaqErr </td>
  *      <td> Generate DAQ errors of type \c P1. The average cluster rate can 
  *           be specified with \c -rate \c \<hz\>, and a fixed or random 
  *           cluster width can be specified with \c -clust \c \<N\> and 
  *           \c -avgwid \c \<nAvg\>, respectively.</td>
  *  </tr>
  *  <tr><td> %Delay </td>
  *      <td> %Delay data by the number of seconds specified in \c P1.</td>
  *  </tr>
  *  <tr><td>\c Design</td>
  *      <td> Use the \c FilterDesign class to define the filter. \c P1
  *           is the sample rate and \c P2 is the \c %FilterDesign 
  *           specification string.</td>
  *  </tr>
  *  <tr><td>\c IfoNoise</td>
  *      <td> Apply an IfoNoise filter to the time series. The IFO noise
  *           filter is implemented in the fourier domain and returns the
  *           noise spectrum for the associated instrument when it operates
  *           on a unit sigma gaussian distribution. The implemented 
  *           instruments are "aLigo", "Geo" and "Virgo". The 
  *           parameterizations were borrowed from Shourov / Patrick Sutton's
  *           omega SRD functions.</td>
  *  </tr>
  *  <tr><td>\c %IfoResponse</td>
  *      <td> Apply an IfoResponse filter to the time series. The IFO 
  *           response is implemented in the fourier domain and uses the 
  *           LscCalib class to fetch the calibration constants from an 
  *           EasyCalibrate compatible xml file. The argument \c P1 is the 
  *           calibration file name.</td>
  *  </tr>
  *  <tr><td>\c %MathOp</td>
  *      <td>Perform element-by-element arithmetic, boolean or logical 
  *          operations on the input time series. P1 [, P3, ...] contains
  *          an operation string ("+", "*", "&", "|", "=", etc.) and 
  *          P2 [, P4, ...] contains a constant argument (see MathOp).</td>
  *  </tr>
  *  <tr><td>\c Recolor</td>
  *      <td>Design a recoloring FIR filter. P1 contains the output PSD
  *          for gaussian input data. If -inPSD is specified, the following 
  *          parameter is the %PSD of the input data. PSDs are specified as
  *          a file path. The file contains two columns with bin-frequency
  *          and amplitude (sqrt(PSD)) </td>
  *  </tr>
  *  <tr><td>\c ReSample</td>
  *      <td> Resample in the fourier domain to a sample rate of \c P1.</td>
  *  </tr>
  *  </table>
  *
  *  \subsection dmtgen_conf_srce Data Source Specifications
  *
  *  Data source specifiers define continuous or transient signal sources,
  *  the outputs of which are filtered and summed to form simulated channels. 
  *  The syntax of the source definition statement is:
  *  \verbatim
    Source <src-name> <src-type>([<param1>=<value1>[,...,<paramN>=<valueN>]]) \
           [-rate <rate> | -time <gps-time>] [-sample <samp-rate>]	      \
	   [-type <data-type>] [-simevent] [-simdata] 
     \endverbatim
  *  The currently available data sources are:
  *
  *  <table>
  *  <tr><td>\c DampedSine(A,F,Q,Phi,Width) </td>
  *      <td> Damped sinusoid[\ref generator::DampedSine]. </td>
  *  </tr>
  *  <tr><td>\c FrameData(Channel,Files) </td>
  *      <td> Read %Channel from data frames in \c Files 
  *           [\ref generator::FrameData]</td>
  *  </tr>
  *  <tr><td>\c GaussBurst(A,Sigma,Width) </td>
  *      <td> White noise modulated by a gaussian envelope 
  *           [\ref generator::GaussBurst].</td>
  *  </tr>
  *  <tr><td>\c Metastable(t0,t1,Off,On) </td>
  *      <td> Return "Off" value for time "t0" followed by "On" value 
  *           for time "t1" [\ref  generator::Metastable].</td>
  *  </tr>
  *  <tr><td>\c Sine(A,F,Phi) </td>
  *      <td> Continuous sinusoid [\ref generator::SineData].</td>
  *  </tr>
  *  <tr><td>\c Sinc(A,F,Width) </td>
  *      <td> Sinc function [\ref generator::Sinc].</td>
  *  </tr>
  *  <tr><td>\c SinGauss(A,F,Q,Phi,Width) </td>
  *      <td> Sinusoid in a gaussian envelope [\ref generator::SinGauss].</td>
  *  </tr>
  *  <tr><td>\c WhiteNoise(A) </td>
  *      <td> Gaussian white noise of amplitude (sigma) \c A 
  *           [\ref generator::WhiteNoise].</td>
  *  </tr>
  *  </table>
  *
  *  Each numeric parameter may be a constant or may be generated randomly 
  *  according to one of several distributions. The currently implemented 
  *  distributions are:
  *
  *  <table>
  *  <tr><td>\c \<const\> </td>
  *      <td> Constant value</td>
  *  </tr>
  *  <tr><td>\c flat(\<min\>,\<max\>) </td>
  *      <td> Flat distribution in the interval \c [\<min\>,\<max\>] </td>
  *  </tr>
  *  <tr><td>\c step(\<min\>,\<max\>,\<step\>) </td>
  *      <td> Step from \c \<min\> to \c \<max\> in increments of 
  *           \c \<step\> </td>
  *  </tr>
  *  <tr><td>\c xstep(\<min\>,\<max\>,\<step\>) </td>
  *      <td> Step exponentially from \c \<min\> to \c \<max\> by factors 
  *           of \c \<step\> </td>
  *  </tr>
  *  <tr><td>\c gauss(\<sigma\>,\<avg\>) </td>
  *      <td> Gaussian distribution with average value \c \<avg\> and 
  *           sigma \c \<sigma\></td>
  *  </tr>
  *  <tr><td>\c power(\<pow\>,\<min\>,\<max\>) </td>
  *      <td> Power law distribution between \c \<min\> and \c \<max\> 
  *           with exponent \c \<pow\> </td>
  *  </tr>
  *  <tr><td>\c exp(\<b\>,\<min\>,\<max\>) </td>
  *      <td> exponential distribution from \c \<min\> to \c \<max\> 
  *           with average exponent \c \<b\> </td>
  *  </tr>
  *  </table>
  *
  *  The additional arguments in the \c Source statment have the following 
  *  meanings:
  *
  *  <table>
  *  <tr><td>\c \<src-name\> </td>
  *      <td> Name by which the source is referred.</td>
  *  </tr>
  *  <tr><td>\c -rate \c \<rate\> </td>
  *      <td> Events are to be generated at random times with average 
  *           rate \c \<rate\> </td>
  *  </tr>
  *  <tr><td>\c -time \c \<gps-time\> </td>
  *      <td> A single event is to be generated at time \c \<gps-time\> </td>
  *  </tr>
  *  <tr><td>\c -type \c \<data-type\> </td>
  *      <td> Samples are to be of type \c \<data-type\>, i.e. one of
  *           {double, float, int, short} </td>
  *  </tr>
  *  <tr><td>\c -sample \c \<samp-rate\> </td>
  *      <td>Data are to be digitized at \c \<samp-rate\> Hz. [16384]</td>
  *  </tr>
  *  <tr><td>\c -simevent </td>
  *      <td> Record generated event parameters.</td>
  *  </tr>
  *  <tr><td>\c -simdata </td>
  *      <td> Record generated source wave-forms.</td>
  *  </tr>
  *  </table>
  *
  *  \subsection dmtgen_conf_chan Channel Specification
  *
  *  %Channel specifications describe a channel and the data sources that 
  *  are to be used to make it up. The syntax of the channel specification is:
  *  \verbatim
     Channel <chan> [-daqchain <filter>] [-sample <rate>] <src1> [... <srcN>]
     \endverbatim
  *  All sources (\c \<src1\>...\<srcN\> ) listed on the channel specification
  *  are summed together and the result is passed through the data acquisition
  *  simulation filter (\c \<filter\> ). Resampling to (\c \<rate\> ) is not 
  *  yet implemented. The channel name (\c \<chan\> ) is used as the channel 
  *  name in the output frames.
  *
  *  \brief DMT data simulation driver program.
  *  \author J. Zweizig
  *  \version 1.4; Last Modified October 30, 2010
  */
class DMTGen {
public:
    /**  DMT signal generation class constructor.
      *  \brief Construct generation driver.
      *  \param argc Number of command line argument words.
      *  \param argv Vector of command line argument strings.
      */
    DMTGen(int argc, const char* argv[]);

    /** Destroy driver program
      * \brief Destructor
      */
    ~DMTGen(void);

    /**  Read a configuration file.
      *  \brief Read configuration.
      *  \param file Configuration file path.
      *  \return True if error occurred in parsing file.
      */
    bool readConfig(const std::string& file);

    /**  Data generation entry
      *  \brief Generate data.
      */
    void generate(void);

    /**  Write a frame of generated data.
      *  \brief Write generated data.
      *  \param t0 Start time.
      *  \param dT Lengt of frame to be written.
      */
    void writeFrame(const Time& t0, Interval dT);

    /**  Test whether configuration was successful
      *  \brief Test configuration.
      *  \return True if successful configuration.
      */
    bool configured(void) const;

private:
    typedef std::map<std::string, const Pipe*> FilterMap;
    typedef FilterMap::const_iterator          const_filt_iter;
    typedef FilterMap::iterator                filt_iter;

private:
    bool                 mConfigured;
    FilterMap            mFilterMap;
    generator::SourceDict mSourceDict;
    std::string          mIdent;
    TrigPar              mDict;
    generator::Generator mGener;
    long                 mRunID;
    std::string          mFilePfx;
    std::string          mOutputDir;
    enum {kFile, kSMP}   mWriteMode;
    FrWriter             mWriter;
    Time                 mStartTime;
    Time                 mEndTime;
    Interval             mDt;
    int                  mDebug;
    int                  mNFramesOut;
    int                  mRealTime;
    SigFlag              mTerm;
};

//======================================  DMTGen inline functions
inline bool 
DMTGen::configured(void) const {
    return mConfigured;
}
