////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                  hMon.hh                                   //
//                                                                            //
//      Generation of time-domain calibrated data and analysis segments       //
//                                                                            //
//                            Shourov K. Chatterji                            //
//                                                                            //
//                                2008-Aug-07                                 //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

// $Id: hMon.hh 5250 2008-09-05 16:39:10Z shourov.chatterji $

#ifndef HMON_HH
#define HMON_HH

#include "DatEnv.hh"
#include "MonServer.hh"
#include "TrigClient.hh"
#include "FrWriter.hh"
#include "Trend.hh"
#include "TSeries.hh"
#include "FixedLenTS.hh"
#include "FSeries.hh"
#include "FSpectrum.hh"
#include "FIRdft.hh"
#include "OperStateCondList.hh"
#include "SegList.hh"
#include "DVecType.hh"
#include "Hanning.hh"
#include "ManagedSegmentList.hh"
#include <string>
#include <vector>
#include <stdexcept>

//
//   Generator of time-domain calibrated data and analyzable segments

/**  @name hMon
  *  @memo Monitor for online generation of time-domain calibrated data.
  *  The hMon monitor produces time-domain calibrated data for use in rapid
  *  online analysis.  It also produces baseline segments for rapid only
  *  analysis that reflect detector lock, science, and injection status, as well
  *  as baseline data quality flags such as missing data, missing calibration,
  *  bad reconstruction, etc.
  *
  *  The calibrated time series output is written to disk as frame files and
  *  also made available for transfer via NDS2.  Segments are published to the
  *  segment database and also written to disk for local as well as remote http
  #  access.
  *
  *  The monitor also makes available the calibrated detector noise spectrum for
  *  viewing by the dmtviewer, as well as the time varying optical gain factors
  *  for each of the calibration lines.  The time varying optical gain factors are
  *  also stored as second and minute trends.
  *
  *  usage:
  *
  *  \begin{verbatim}
        hMon
     \end{verbatim}
  *
  *  @author Shourov Chatterji; shourov@ligo.caltech.edu
  *  @version 1.1; Last modified 2008-Aug-06
  */
//{
//}

class hMon : public DatEnv, MonServer, TrigClient
{

public:

  hMon(int argc, const char *argv[]);		  // Constructor
  ~hMon(void);					  // Destructor
  void ProcessData(void);			  // Process 1 data stride
  void Attention(void);				  // Handle Message signal.

private:

  FrWriter frameWriter;                           // Frame Writer
  Trend alphaMinuteTrend;                         // Trend writer

  int verboseLevel;                               // Verbose level of log output

  std::string configurationFile;                  // Name of configuration file

  std::string processName;                        // Executable name
  std::string detectorName;                       // Detector name
  std::string siteName;                           // Site character identifier
  std::string monitorName;                        // Detector specific monitor name

  std::string sensingChannelName;                 // Sensing channel name
  std::string errorChannelName;                   // Error channel name
  std::string controlChannelName;                 // Control channel name
  std::string excitationChannelName;              // Excitation channel name

  FixedLenTS sensingChannelData;                  // Sensing channel data
  FixedLenTS errorChannelData;                    // Error channel data
  FixedLenTS controlChannelData;                  // Control channel data
  FixedLenTS excitationChannelData;               // Excitation channel data

  TSeries calibratedData;                         // Calibrated time series data

  std::string sensingFilterFile;                  // Sensing filter coefficients file
  std::string servoFilterFile;                    // Servo filter coefficients file
  std::string actuationFilterFile;                // Actuation filter coefficients file
  std::string lowPassFilterFile;                  // Low pass filter coefficients file
  std::string highPassFilterFile;                 // High pass filter coefficients file

  FIRdft sensingFilter;                           // Sensing filter coefficients file
  FIRdft servoFilter;                             // Servo filter coefficients file
  FIRdft actuationFilter;                         // Actuation filter coefficients file
  FIRdft lowPassFilter;                           // Low pass filter coefficients file
  FIRdft highPassFilter;                          // High pass filter coefficients file

  std::vector<double> excitationFrequencies;      // Calibration line frequencies
  std::vector<double> excitationAmplitudes;       // Calibration line amplitudes

  std::string frameType;                          // Output frame file type
  std::string frameDirectory;                     // Output frame file directory
  std::string frameChannelName;                   // Output frame file channel name
  std::string framePrecision;                     // Output frame file precision
  int frameEpochDuration;                         // Output frame file directory duration
  FrVectRef::compression_mode frameCompression;   // Output frame file compression mode

  int sampleRate;                                 // Sample frequency of input data
  int upSampleFactor;                             // Sample frequency factor for filtering
  int frameSampleRate;                            // Sample frequency of output data

  Interval sampleInterval;                        // Inverse of sample rate
  Interval filterDuration;                        // FIR filter duration
  Interval filterDelay;                           // FIR filter delay
  Interval unitStride;                            // Fundamental stride
  Interval frameStride;                           // Frame file stride
  Interval segmentStride;                         // Segment publishing stride
  Interval latencyThreshold;                      // Maximum allowable latency

  Time currentReadTime;                           // Start time of most recently read data
  Time lastReadTime;                              // Stop time of previously read data
  Time calibratedStartTime;                       // Start time of calibirated data stride
  Time lastCalibratedTime;			  // Stop time of most recently calibrated data
  Time realTime;                                  // Real time from system clock

  float alpha;                                    // Optical gain factor
  FixedLenTS alphaTimeSeries;                     // Optical gain factor time

  OperStateCondList detectorStates;               // List of detector states to watch
  std::string detectorStatesFile;                 // Definition file for detector states

  std::string segmentDirectory;                   // Output segment directory
  Interval segmentListDuration;                   // Duration of segment lists to publish

  int segmentVersion;                             // Version number for segment publication

  ManagedSegmentListMap segmentLists;		  // Container for segment lists

  double spectrumResolution;                      // Frequency resolution of spectrum
  Hanning spectrumWindow;                         // Window for computing spectrum
  Interval spectrumDecayTime;                     // Time scale for averaging spectrum
  FSpectrum currentSpectrum;                      // Current calibrated power spectrum
  FSpectrum averageSpectrum;                      // Average calibrated power spectrum
  double spectrumWeightingFactor;                 // Weighting factor for averaging spectra
  float horizonDistance;                          // Range to optimally oriented fiducial source
  float senseMonRange;                            // Range to randomly oriented fiducial source
  float figureOfMerit;                            // LIGO design figure of merit

  Interval serverDuration;                        // Duration of dmtviewer data products

  FixedLenTS alphaHistory;                        // Optical gain factor history
  FixedLenTS horizonDistanceHistory;              // SenseMon range history
  FixedLenTS senseMonRangeHistory;                // Horizon distance history
  FixedLenTS figureOfMeritHistory;                // Figure of merit history

  // Time stamp for log file output
  std::string logStamp(int verboseThreshold);

  // Read command line arguments
  void readArguments(int argc, const char *argv[]);

  // Set default parameter values
  void defaultParameters(void);

  // Read parameter values from configuration file
  void readConfiguration(void);

  // Determine derived parameter values
  void derivedParameters(void);

  // Test for invalid parameter values
  void validateParameters(void);

  // Initialize data accessors and containers
  void initializeData(void);

  // Initialize calibrated spectrum
  void initializeSpectrum(void);

  // Initialize dmtviewer server
  void initializeServer(void);

  // Initialize second and minute trends
  void initializeTrends(void);

  // Initialize filter coefficients
  void initializeFilters(void);

  // Read filter file
  void readFilterFile(const std::string& filterFile, FIRdft& filter,
		      int filterRate);

  // Initialize detector state definitions
  void initializeStates(void);

  // Initialize segments
  void initializeSegments(void);

  // Read next unit stride of data
  void readData(void);

  // Read data from a single channel
  void readChannel(const std::string& channelName, FixedLenTS& channelData);

  // Test for excessive latency
  void testLatency(void);

  // Update segment lists
  void updateSegments(void);

  // Finalize segment lists
  void finalizeSegments(void);

  // Reset filter history
  void resetFilters(void);

  // Reset spectrogram computation
  void resetSpectrum(void);

  // Test for complete frame stride
  bool isFrameStride(void);

  // Compute optical gain factors
  void computeFactors(void);

  // Compute time-domain calibrated data
  void calibrateData(void);

  // Compute calibrated spectrum
  void computeSpectrum(void);

  // Compute range to fiducial source
  void computeSensitivity(void);

  // Publish calibrated data frames
  void publishFrames(void);

  // Publish segment lists
  void publishSegments(void);

  // Write segments lists to file
  void writeSegments(std::string segmentListName);

  // Send segments to database
  void sendSegments(std::string segmentListName);

  // Serve data for dmtviewer
  void updateServer(void);

  // Update minute and second trends
  void updateTrends(void);

};

// Warning class for non-fatal exceptions
class hMonWarning : public std::runtime_error
{
public:
  hMonWarning(const std::string& message) : std::runtime_error(message) {}
};

// Error class for fatal exceptions
class hMonError : public std::runtime_error
{
public:
  hMonError(const std::string& message) : std::runtime_error(message) {}
};

#endif // HMON_HH
